#include "constants.h"
#include "defaultplayercontroller.h"
#include "input.h"
#include "internal.h"
#include "internal/input.h"
#include "internal/physics.h"
#include "internal/ui.h"
#include "noclipplayercontroller.h"
#include "types.h"
#include "ui.h"
#include "uielements.h"
#include <GLFW/glfw3.h>

namespace kek::Defaults {

static DefaultPlayerController *defaultController;
static NoclipPlayerController *noclipController;

static KeyBinding
	keyOptions,
	keyToggleCursorMode,
	keyToggleNoclip,
	keyExit;

static ButtonElement *options;

static void defaultPeriodic(PeriodicEvent event, void *data) {
	kekData.player->controller->update();
}

static void doUIMouseHover(double x, double y) {
	if(Input::getMouseButtonState(GLFWMouseButton::LEFT) == GLFWMouseButtonState::PRESSED) return;

	IntPoint2 mousePosition = IntPoint2((int) x, (int) y);
	UIElement *hoveredElement = UI::findElementAt(mousePosition);

	for(UIElement *element : kekData.ui->elements) {
		if(element->hovering && element != hoveredElement) {
			element->hoverExitAll();
		}
	}

	if(hoveredElement) {
		IntPoint2 childPos = hoveredElement->getScreenPosition();
		hoveredElement->hoverAll(mousePosition - childPos, mousePosition);
	}
}

static void stopUIMouseDrag() {
	if(!kekData.ui->draggedElement) return;

	double x, y;
	glfwGetCursorPos(kekData.window, &x, &y);
	IntPoint2 mousePosition = IntPoint2((int) x, (int) y);

	IntPoint2 childPos = kekData.ui->draggedElement->getScreenPosition();
	kekData.ui->draggedElement->dragExit(mousePosition - childPos, mousePosition);
	kekData.ui->draggedElement = nullptr;
}

static void doUIMouseDrag(double x, double y) {
	IntPoint2 mousePosition = IntPoint2((int) x, (int) y);

	if(Input::getMouseButtonState(GLFWMouseButton::LEFT) != GLFWMouseButtonState::PRESSED) {
		kekData.input->mouseDown = false;
		stopUIMouseDrag();
		return;
	}

	if(!kekData.input->mouseDown) {
		kekData.input->mouseDownPosition = mousePosition;
	}

	if(!kekData.ui->draggedElement) {
		IntPoint2 mouseDownPosition = kekData.input->mouseDownPosition;
		int deltaX = (int) x - mouseDownPosition.x;
		int deltaY = (int) y - mouseDownPosition.y;
		if(deltaX * deltaX + deltaY * deltaY >= KEK_UI_MIN_DRAG_DISTANCE_SQUARED) {
			// Initiate drag
			UIElement *draggedElement = UI::findElementAt(mouseDownPosition);

			if(draggedElement) {
				IntPoint2 childPos = draggedElement->getScreenPosition();
				draggedElement->dragEnterAll(mouseDownPosition - childPos, mouseDownPosition);
			}
		}
	} else {
		// Element is being dragged
		IntPoint2 childPos = kekData.ui->draggedElement->getScreenPosition();
		kekData.ui->draggedElement->drag(mousePosition - childPos, mousePosition);
	}

	kekData.input->mouseDown = true;
}

static void doUIMouseClick(double x, double y, GLFWMouseButton button) {
	IntPoint2 mousePosition = IntPoint2((int) x, (int) y);
	UIElement *clickedElement = UI::findElementAt(mousePosition);

	if(clickedElement) {
		IntPoint2 childPos = clickedElement->getScreenPosition();
		IntPoint2 clickedAt = mousePosition - childPos;

		if(button == GLFWMouseButton::LEFT) {
			if(!clickedElement->focusEnterAll(clickedAt, IntPoint2((int) x, (int) y))) {
				UI::unfocusElement(kekData.ui->focusedElement);
			}
		}

		clickedElement->clickAll(clickedAt, IntPoint2((int) x, (int) y), button);
	} else {
		UI::unfocusElement(kekData.ui->focusedElement);
	}
}

static void defaultKeyCallback(KeyEvent event, void *data) {
	if(event.key == Input::getKeyBinding(keyExit).key && event.action == GLFW_PRESS) {
		glfwSetWindowShouldClose(event.window, true);
	}

	if(event.key == Input::getKeyBinding(keyOptions).key && event.action == GLFW_PRESS) {
		options->visible = !options->visible;
	}

	if(event.key == Input::getKeyBinding(keyToggleCursorMode).key && event.action == GLFW_PRESS) {
		if(Input::getCursorMode() == GLFWCursorMode::CAPTURE) {
			Input::setCursorMode(GLFWCursorMode::FREE);
			double x, y;
			glfwGetCursorPos(kekData.window, &x, &y);
			doUIMouseHover(x, y);
			doUIMouseDrag(x, y);
		} else {
			Input::setCursorMode(GLFWCursorMode::CAPTURE);

			for(auto e : UI::getElements()) {
				if(e->hovering) e->hoverExitAll();
			}
		}
	}

	if(event.key == Input::getKeyBinding(keyToggleNoclip).key && event.action == GLFW_PRESS) {
		kekData.player->noclip = !kekData.player->noclip;
		kekData.player->controller = kekData.player->noclip ? (PlayerController *) noclipController : (PlayerController *) defaultController;
		if(kekData.player->noclip) {
			kekData.physics->world->removeRigidBody(kekData.player->physics->body);
		} else {
			kekData.physics->world->addRigidBody(kekData.player->physics->body);
		}
	}
}

static void defaultMouseCallback(MouseEvent event, void *data) {
	static bool firstMouse = true;
	static double lastX = 0, lastY = 0;
	if(firstMouse) {
		lastX = event.x;
		lastY = event.y;
		firstMouse = false;
	}

	switch(Input::getCursorMode()) {
		case GLFWCursorMode::CAPTURE: {
			float xoff = lastX - event.x;
			float yoff = lastY - event.y;

			xoff *= 0.1f;
			yoff *= 0.1f;

			kekData.activeCamera->rotateYaw(xoff);
			kekData.activeCamera->rotatePitch(yoff);
			break;
		}
		case GLFWCursorMode::FREE:
		case GLFWCursorMode::HIDDEN: {
			doUIMouseHover(event.x, event.y);
			doUIMouseDrag(event.x, event.y);
			break;
		}
	}

	lastX = event.x;
	lastY = event.y;
}

static void defaultMouseButtonCallback(MouseButtonEvent event, void *data) {
	switch(Input::getCursorMode()) {
		case GLFWCursorMode::CAPTURE: {
			// TODO
			break;
		}
		case GLFWCursorMode::FREE:
		case GLFWCursorMode::HIDDEN: {
			switch(event.action) {
				case GLFW_PRESS:
					break;
				case GLFW_RELEASE: {
					kekData.input->mouseDown = false;
					stopUIMouseDrag();

					double x, y;
					glfwGetCursorPos(event.window, &x, &y);
					doUIMouseClick(x, y, (GLFWMouseButton) event.button);
				}
			}

			break;
		}
	}
}

void init() {
	defaultController = new DefaultPlayerController();
	noclipController = new NoclipPlayerController();
	defaultController->keyForward = noclipController->keyForward = Input::createKeyBinding("Forward", GLFW_KEY_W);
	defaultController->keyBackward = noclipController->keyBackward = Input::createKeyBinding("Backward", GLFW_KEY_S);
	defaultController->keyLeft = noclipController->keyLeft = Input::createKeyBinding("Left", GLFW_KEY_A);
	defaultController->keyRight = noclipController->keyRight = Input::createKeyBinding("Right", GLFW_KEY_D);
	defaultController->keyJump = noclipController->keyUp = Input::createKeyBinding("Jump", GLFW_KEY_SPACE);
	noclipController->keyDown = Input::createKeyBinding("Down", GLFW_KEY_LEFT_CONTROL);
	keyOptions = Input::createKeyBinding("Options", GLFW_KEY_Q);
	keyToggleCursorMode = Input::createKeyBinding("Toggle Cursor Mode", GLFW_KEY_TAB);
	keyToggleNoclip = Input::createKeyBinding("Toggle Noclip", GLFW_KEY_N);
	keyExit = Input::createKeyBinding("Exit", GLFW_KEY_ESCAPE);

	Input::addPeriodicCallback(PeriodicCallback(defaultPeriodic, nullptr));
	Input::addKeyListener(KeyCallback(defaultKeyCallback, nullptr));
	Input::addMouseListener(MouseCallback(defaultMouseCallback, nullptr));
	Input::addMouseButtonListener(MouseButtonCallback(defaultMouseButtonCallback, nullptr));

	options = new ButtonElement(uiPx(0), uiPx(100), uiPx(100), uiPx(50));
	// UI::addElement(options);

	kekData.player = new Player();
	btVector3 positions[2];
	positions[0] = btVector3(0, -KEK_PLAYER_HEIGHT / 2 + KEK_PLAYER_RADIUS, 0);
	positions[1] = btVector3(0, KEK_PLAYER_HEIGHT / 2 + KEK_PLAYER_RADIUS, 0);
	float radii[2] = {KEK_PLAYER_RADIUS, KEK_PLAYER_RADIUS};
	btCollisionShape *shape = new btMultiSphereShape(positions, radii, 2);
	btRigidBody *body = new btRigidBody(0, nullptr, shape);
	body->setCollisionFlags(body->getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT);
	body->setFriction(3);
	body->setActivationState(DISABLE_DEACTIVATION);
	kekData.physics->world->addRigidBody(body);

	btTransform from;
	from.setIdentity();
	from.setOrigin(btVector3(0, 0, 0));

	btTransform to;
	to.setIdentity();
	to.setOrigin(btVector3(1, 1, 1));

	btCollisionWorld::ClosestConvexResultCallback cb = btCollisionWorld::ClosestConvexResultCallback(from.getOrigin(), to.getOrigin());
	kekData.physics->world->convexSweepTest((btConvexShape *) shape, from, to, cb);

	kekData.player->physics = new PlayerPhysicsObjectData();
	kekData.player->physics->body = body;
	kekData.player->moveTo(glm::vec3(5, 10, 0));
	kekData.player->controller = defaultController;
}

void destroy() {
}

}
